/*
 * Segment 2 (KEx),  SWE 2.1 LMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lms.versand.kodierer;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.lms.versand.ftpclient.LMSFTPClient;
import de.bsvrz.kex.lms.vew.LMSFTPListener;
import de.bsvrz.kex.lms.vew.LMSKommunikationsstatus;
import de.bsvrz.kex.lms.vew.LMSParameter;
import de.bsvrz.kex.lms.vew.LMSProperties;
import de.bsvrz.kex.lms.vew.LmsMsg;
import de.bsvrz.kex.lms.vew.MeldungsStatus;
import de.bsvrz.kex.lms.vew.RDSException;
import de.bsvrz.kex.lms.vew.RDSMeldung;
import de.bsvrz.kex.lms.vew.XMLTools;
import de.bsvrz.kex.lms.vew.daten.DavDatenModell;
import de.bsvrz.kex.lms.vew.daten.RDSStatus.RDS_StatusEN;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Modul Meldungskodierer.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */

public class Meldungskodierer extends Thread
implements ClientReceiverInterface, ClientSenderInterface, LMSFTPListener, InvalidationListener {

	/**
	 * Listener f&uuml;r &Auml;nderungen der Menge der RDS-Meldungen.
	 */
	class MengeListener implements MutableSetChangeListener {

		@Override
		public void update(final MutableSet set, final SystemObject[] addedObjects,
				final SystemObject[] removedObjects) {
			for (final SystemObject so : addedObjects) {
				debug.info("Die neue Meldung '" + so.getPid() + "' wird hinzugefügt!");
				try {
					hinzufuegenMeldung(so);
				} catch (final RDSException e) {
					debug.error(e.getMessage());
				}
			}

			for (final SystemObject so : removedObjects) {
				debug.info("Die Meldung '" + so.getPid() + "' wird aus der internen Liste entfernt!");
				entfernenMeldung(so);
			}
		}
	}

	/**
	 * Attributgruppe der Parameter.
	 */
	private static final String ATTRIBUTGRUPPE_PARAMETER = "atg.rdsKommunikationsParameter";

	/**
	 * Wurzelelement der TIC-Info Dateien.
	 */
	private static final String XML_ELEMENT_TIC_ROOT = "IFS";

	/**
	 * DebugLogger für Debug-Ausgaben.
	 */
	final Debug debug = Debug.getLogger();

	/**
	 * DaV Interface.
	 */
	private ClientDavInterface dav = null;

	// private final String DAV_TYP = "typ.landesMeldeStelle";

	/**
	 * Datenmodell der Konfiguration.
	 */
	private DataModel config = null;

	/**
	 * Objektreferenz auf das Objekt vom Typ "typ.landesMeldeStelle".
	 */
	private ConfigurationObject lms = null;

	/**
	 * Die Liste der zu versendenden Meldungen.
	 */
	Map<String, RDSMeldung> meldungen = new HashMap<String, RDSMeldung>();

	/**
	 * Map zur Zuordnung von Dateien zu Meldungen.
	 */
	private final Map<File, List<RDSMeldung>> dateiMap = new HashMap<File, List<RDSMeldung>>();

	/**
	 * FTP-Client Thread.
	 */
	LMSFTPClient ftpclient;

	/**
	 * Diese Klasse versendet Betriebsmeldungen.
	 */
	protected MessageSender nachrichtenSender = null;

	/** Datenbeschreibung, mit der Parameter empfangen werden. */
	private DataDescription paramEmpfang = null;

	/** Datenbeschreibung, mit der Parameter versendet werden. */
	private DataDescription paramVersand = null;

	/** Parameter. */
	private final LMSParameter lmsParameter = new LMSParameter();

	/** Der Kommunikationsstatus mit der Landesmeldestelle. */
	private LMSKommunikationsstatus kommunikationsStatus;

	/** Lock-Objekt. */
	private final Object runLock = new Object();

	/** Run-Flag. */
	boolean run = true;

	/**
	 * Konstruktor.
	 *
	 * @param dav
	 *            Datenverteilerverbindung
	 * @param lms
	 *            das die Landesmeldestelle repr&auml;sentierende Systemobjekt
	 */
	public Meldungskodierer(final ClientDavInterface dav, final ConfigurationObject lms) {
		super();
		this.dav = dav;
		this.lms = lms;
		config = dav.getDataModel();
		this.nachrichtenSender = MessageSender.getInstance();
		this.setName(this.getClass().getSimpleName());
		// this.nachrichtenSender.init(dav, "LMS-Meldungsversand",
		// "LMS-Meldungsversand");

		DavDatenModell.getInstanz().setVerbindung((ClientDavConnection) dav);
		this.nachrichtenSender.setApplicationLabel("LMS-Meldungen");

		// selbst starten
		start();
	}

	/**
	 * Bildet den Dateinamen f&uuml;r eine zu versendende Datei.
	 *
	 * @return Dateiname
	 */
	private String bildeDateinamen() {
		return LMSProperties.getInstance().getProperty(LMSProperties.PROP_KEY_VERSANDDATEIPREFIX) + "_"
				+ System.currentTimeMillis() + ".xml";
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		debug.finest("dataRequest");

	}

	/**
	 * Meldet alle ben&ouml;tigten Daten am Datenverteiler an.
	 */
	protected void datenAnmeldenDaV() {
		final DataDescription dataDescription = null;
		SystemObject[] objekte = null;
		AttributeGroup attributeGroup = null;
		Aspect aspect = null;

		// Anmeldungen zum Senden und Empfangen.
		try {
			//
			// Anmeldungen zum Empfang
			//
			// Kommunikationsparameter
			objekte = new SystemObject[] { lms };
			attributeGroup = config.getAttributeGroup(Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER);
			aspect = config.getAspect("asp.parameterSoll");
			paramEmpfang = new DataDescription(attributeGroup, aspect);

			// aktuelle Parameter lesen
			final ResultData resultdaten = dav.getData(lms, paramEmpfang, 0);
			lmsParameter.aktualisieren(resultdaten.getData());

			debug.finer("Anmeldung zum Empfang von Daten beim DaV als SENKE :" + "(Objekt: " + lms.getName()
			+ ") (ATG: " + Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER + ") (ASP: " + "asp.parameterSoll"
			+ ") (OT: " + lms.getType().getName() + ")");

			dav.subscribeReceiver(this, objekte, paramEmpfang, new ReceiveOptions(false), // keine
					// Deltas,
					// keine
					// nachgelieferten Daten
					ReceiverRole.receiver());

			// Menge RDS-Meldungen anmelden
			final MutableSet meldungsmenge = lms.getMutableSet("RDSMeldungen");
			debug.info("Menge 'RDSMeldungen' am Objekt " + lms + ": " + meldungsmenge);

			// meldungsmenge.addChangeListener(new MengeListener());
			for (final SystemObject meldung : meldungsmenge.getElements()) {
				debug.info("Die aktuelle RDS-Meldung " + meldung + " wird initialisiert");
				try {
					hinzufuegenMeldung(meldung);
				} catch (final RDSException e) {
					debug.error(e.getMessage());
				}
			}
			meldungsmenge.addChangeListener(new MengeListener());

			//
			// Anmeldungen zum Senden
			//
			// Kommunikationsparameter
			objekte = new SystemObject[] { lms };
			attributeGroup = config.getAttributeGroup(Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER);
			aspect = config.getAspect("asp.parameterVorgabe");
			paramVersand = new DataDescription(attributeGroup, aspect);

			debug.finer("Anmeldung zum Senden von Daten an den DaV als SENDER :" + "(Objekt: " + lms.getName()
			+ ") (ATG: " + Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER + ") (ASP: " + "asp.parameterVorgabe"
			+ ") (OT: " + lms.getType().getName() + ")");

			dav.subscribeSender(this, objekte, paramVersand, SenderRole.sender());

			// IvalidationListener
			final DynamicObjectType typ = (DynamicObjectType) config.getType("typ.rdsMeldung");
			// typ.addObjectCreationListener(privateListener);
			typ.addInvalidationListener(this);

		} catch (final OneSubscriptionPerSendData e) {
			debug.error("Daten wurden bereits als Quelle angemeldet: " + dataDescription);
			throw new RuntimeException(e);
		} catch (final DataNotSubscribedException e) {
			// TBD: Die Exception ist wiedersprüchlich dokumentiert --> Roland
			debug.error("Schreibanmeldung fehlgeschlagen: " + dataDescription);
			System.exit(0);
			throw new RuntimeException(e);
		}
	}

	@Override
	public void empfangen(final List<File> dateien) {
		// das sollte hier nicht auftreten
		debug.error("unerwartetes Listener-Event 'empfangen'");
	}

	/**
	 * Entfernt eine Meldung aus der Meldungsliste.
	 *
	 * @param so
	 *            zugeordnetes Systemobjekt
	 */
	void entfernenMeldung(final SystemObject so) {
		synchronized (meldungen) {
			final RDSMeldung rdsMeldung = meldungen.get(so.getPid());
			if (rdsMeldung != null) {
				try {
					rdsMeldung.abmeldenDavVersand(dav);
				} catch (final RDSException e) {
					debug.error("abmeldenDavVersand: " + e.getMessage());
				}
				meldungen.remove(so.getPid());
			}
		}
	}

	/**
	 * F&uuml;get eine Meldung zur Meldungsliste hinzu.
	 *
	 * @param objekt
	 *            Systemobjekt der Meldung
	 * @throws RDSException
	 *             bei Ausnahmen
	 */
	public void hinzufuegenMeldung(final SystemObject objekt) throws RDSException {
		synchronized (meldungen) {
			final RDSMeldung meldung = new RDSMeldung(objekt);
			meldung.anmeldenDavVersand(dav);
			meldungen.put(meldung.getPID(), meldung);
		}
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		debug.finest("isRequestSupported");
		return false;
	}

	@Override
	public void kommunikationsStatus(final LMSKommunikationsstatus status) {
		final boolean changed = kommunikationsStatus != status;

		kommunikationsStatus = status;

		if (kommunikationsStatus == LMSKommunikationsstatus.GESTOERT) {
			synchronized (dateiMap) {
				synchronized (meldungen) {
					for (final File datei : dateiMap.keySet()) {
						for (final RDSMeldung meldung : dateiMap.get(datei)) {
							debug.fine("Meldung " + meldung + " konnte nicht versendet werden");
							meldung.setStatus(MeldungsStatus.GUELTIG);
						}
					}
				}
				dateiMap.clear();
			}
		}

		if (changed) {
			if (kommunikationsStatus == LMSKommunikationsstatus.GESTOERT) {
				LogTools.log(debug, LmsMsg.WarningStoerungAusgehend, lms.getName());
			}
			if (kommunikationsStatus == LMSKommunikationsstatus.NICHT_GESTOERT) {
				LogTools.log(debug, LmsMsg.InfoStatusAusgehendOK, lms.getName());
			}
		}
	}

	/**
	 * Aktualisiert die laufende Nummer der Kommunikationsparameter.
	 */
	public void laufendeNummerAktualisieren() {
		try {
			ResultData resultdaten = dav.getData(lms, paramEmpfang, 0);
			final Data neueDaten = resultdaten.getData().createModifiableCopy();
			lmsParameter.setLaufendeNummer(neueDaten.getUnscaledValue("LaufendeNummer").longValue() + 1);
			neueDaten.getUnscaledValue("LaufendeNummer").set(lmsParameter.getLaufendeNummer());
			resultdaten = new ResultData(lms, paramVersand, System.currentTimeMillis(), neueDaten);
			dav.sendData(resultdaten);

			// warte bis die Parameter aktualisiert sind
			synchronized (this) {
				try {
					this.wait(1000);
				} catch (final InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		} catch (final DataNotSubscribedException e) {
			LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
		} catch (final SendSubscriptionNotConfirmed e) {
			LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
		}
	}

	@Override
	public void run() {

		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
			@Override
			public void uncaughtException(final Thread t, final Throwable e) {
				debug.error("Meldungskodierer beendet (" + e + ")"); //$NON-NLS-1$
				debug.error(LoggerTools.getStackTrace(e));
				run = false;
				if ((ftpclient != null) && ftpclient.isAlive()) {
					ftpclient.stopp();
					try {
						ftpclient.join();
					} catch (final InterruptedException e1) {
						// TODO Auto-generated catch block
						// e1.printStackTrace();
					}
				}
			}
		});

		synchronized (this) {
			synchronized (runLock) {
				datenAnmeldenDaV();
				ftpclient = new LMSFTPClient(lmsParameter, true);
				ftpclient.addListener(this);
			}

			while (run) {
				try {
					wait(1000);
					try {
						ueberpruefeMeldungen();
					} catch (final RDSException e) {
						debug.error(e.getMessage());
					}
				} catch (final InterruptedException e) {
					/** */
				}
			}

			if (ftpclient != null) {
				ftpclient.stopp();
			}
		}
	}

	/**
	 * Stoppt den Thread.
	 */
	public void stopp() {
		this.run = false;
		this.interrupt();
	}

	/**
	 * &Uuml;berpro&uuml;ft alle Meldungen auf einen notwendigen Versand.
	 *
	 * @throws RDSException
	 *             bei Ausnahmen
	 */
	private void ueberpruefeMeldungen() throws RDSException {
		final List<RDSMeldung> loeschen = new ArrayList<RDSMeldung>();
		final List<RDSMeldung> versenden = new ArrayList<RDSMeldung>();

		synchronized (meldungen) {
			// boolean versand = false;

			for (final RDSMeldung meldung : meldungen.values()) {
				try {
					if (!meldung.isDavDatenOK() || (meldung.getStatus() == MeldungsStatus.FEHLERHAFT)) {
						continue;
					}

					// die Meldung ist aufgehoben und versendet
					if (((meldung.getRDSStatus() == RDS_StatusEN.Aufhebung)
							|| (meldung.getRDSStatus() == RDS_StatusEN.Loeschung))
							&& (meldung.getStatus() == MeldungsStatus.VERSENDET)) {
						loeschen.add(meldung);
						debug.info("Die RDS-Meldung: " + meldung + " wird aus der internen Liste gelöscht");
						continue;
					}

					// Die Meldung wird unter folgenden Bedingungen versendet:
					// - die Daten sind aktualisiert
					// oder
					// - der letzte Versand liegt länger als die parametrierte
					// Wiederholungszeit zurück
					// und
					// - die Meldung ist noch gültig
					// oder
					// - die Meldung wird implizit aufgehoben und wurde bereits
					// versendet
					if (meldung.getStatus() != MeldungsStatus.VERSENDEN) {
						if ((((meldung.getStatus() == MeldungsStatus.AKTUALISIERT)
								|| ((System.currentTimeMillis() - meldung.getLetzterVersand()) > lmsParameter
										.getSendeWiederholung()))
										&& meldung.istZuVerarbeiten())
										|| (meldung.istImplizitAufgehoben()
												&& (meldung.getStatus() == MeldungsStatus.VERSENDET))) {

							meldung.setStatus(MeldungsStatus.VERSENDEN);
							versenden.add(meldung);
							continue;
						} else {
							if (!meldung.istZeitGueltig()) {
								debug.info("Die RDS-Meldung: " + meldung
										+ " ist nicht (mehr) gültig (Ablaufzeit<aktuelle Zeit) und wird deshalb nicht versendet");

								// (erstmal) nicht mehr bearbeiten
								meldung.setStatus(MeldungsStatus.FEHLERHAFT);
							}

							// Das erzeugt ständig Ausgaben
							// else {
							// debug.info("Die RDS-Meldung: " + meldung
							// +
							// " erfüllt die Bedingungen für einen Versand nicht
							// und wird deshalb nicht versendet");
							// }

						}
					}
				} catch (final Exception e) {
					debug.error(e.getMessage());
				}
			}

			if (versenden.size() > 0) {
				versendeMeldungen(new File(bildeDateinamen()), versenden);
			}
		}

		for (final RDSMeldung meldung : loeschen) {
			entfernenMeldung(meldung.getSystemObject());
		}
	}

	@Override
	public void update(final ResultData[] results) {
		for (final ResultData datensatz : results) {
			if (datensatz.hasData()) {
				if (datensatz.getDataDescription().equals(paramEmpfang)) {
					debug.info("Parameterdatensatz erhalten für ",
							datensatz.getObject() + ", " + datensatz.getDataDescription().getAttributeGroup());
					lmsParameter.aktualisieren(datensatz.getData());
					synchronized (this) {
						this.notify();
					}
				}
			}
		}
	}

	/**
	 * Versendet eine Liste von Meldungen.
	 *
	 * @param datei
	 *            Date zum Versand der Meldungen
	 * @param neuemeldungen
	 *            zu versendende Meldungen
	 * @throws RDSException
	 *             bei Ausnahmen
	 */
	public void versendeMeldungen(final File datei, final List<RDSMeldung> neuemeldungen) throws RDSException {
		synchronized (runLock) {
			try {
				boolean senden = false;
				final List<RDSMeldung> versendet = new ArrayList<RDSMeldung>();

				final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				final Document doc = factory.newDocumentBuilder().newDocument();

				final Element root = doc.createElement(Meldungskodierer.XML_ELEMENT_TIC_ROOT);
				doc.appendChild(root);
				root.setAttribute("FileFormatVersion", "2.15.3.1");
				root.setAttribute("FGT", XMLTools.getTICDateTime(System.currentTimeMillis()));
				root.setAttribute("Stock", "1");

				for (final RDSMeldung meldung : neuemeldungen) {
					if (meldung.getStatus() == MeldungsStatus.VERSENDEN) {
						debug.info("Die RDS-Meldung: " + meldung + " wird versendet");
						try {
							laufendeNummerAktualisieren();
							meldung.datenvervollstaendigen(lmsParameter);
							meldung.erzeugeDocumentElement(doc, root);
							meldung.setLetzterVersand(System.currentTimeMillis());
							versendet.add(meldung);
							senden = true;
						} catch (final RDSException e) {
							LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
							meldung.setStatus(MeldungsStatus.FEHLERHAFT);
						}
					}
				}

				if (senden) {
					XMLTools.schreibeXMLDatei(doc, datei);

					dateiMap.put(datei, versendet);
					ftpclient.setVersandDatei(datei);
					synchronized (ftpclient) {
						ftpclient.notify();
					}
					ftpclient.interrupt();
				}
			} catch (final Exception e) {
				e.printStackTrace();
				throw new RDSException(e.getMessage());
			}
		}
	}

	@Override
	public void versendet(final List<File> dateien) {
		synchronized (dateiMap) {
			synchronized (meldungen) {
				for (final File datei : dateien) {
					if (dateiMap.containsKey(datei)) {
						for (final RDSMeldung meldung : dateiMap.get(datei)) {
							debug.fine("Meldung " + meldung + " erfolgreich versendet");
							meldung.setStatus(MeldungsStatus.VERSENDET);
							meldung.publiziereVersendet(dav);
							meldung.setVersendet(true);
						}
					}
					LogTools.log(debug, LmsMsg.FineDateiGeloescht, datei);
					datei.delete();
					dateiMap.remove(datei);
				}
			}
		}
	}

	@Override
	public void invalidObject(final DynamicObject dynamicObject) {
		debug.info("Meldung: " + dynamicObject.getPid() + " wurde invalidiert!");
		entfernenMeldung(dynamicObject);
	}

}
