/*
 * Segment 2 (KEx),  SWE 2.1 LMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lms.vew;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.resource.Configuration;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.lms.MeldungenLoeschen;
import de.bsvrz.kex.lms.vew.daten.DavDatenModell;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;

/**
 * Erzeugt eine RDS-Meldung im Datenverteiler.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class ErzeugeMeldung implements ClientSenderInterface {
	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(MeldungenLoeschen.class.getName());

	private static ClientDavConnection verbindung;

	@BeforeClass
	public static void beforeClass() throws Exception {
		final ObjektFactory factory;

		final Configuration konfig;
		final Level level;

		konfig = Configuration.getConfiguration();
		level = Level.parse(konfig.getString("debug.level"));
		LoggerTools.setLoggerLevel(level);
		ErzeugeMeldung.log.setLevel(level);

		ErzeugeMeldung.log.info("Stelle Verbindung zum Datenverteiler her ...");
		ErzeugeMeldung.verbindung = new ClientDavConnection();
		ErzeugeMeldung.verbindung.getClientDavParameters().setDavCommunicationAddress(konfig.getString("dav.host"));
		ErzeugeMeldung.verbindung.getClientDavParameters().setDavCommunicationSubAddress(konfig.getInt("dav.port"));
		ErzeugeMeldung.verbindung.connect();
		ErzeugeMeldung.verbindung.login(konfig.getString("dav.benutzer"), konfig.getString("dav.kennwort"));

		factory = ObjektFactory.getInstanz();
		factory.setVerbindung(ErzeugeMeldung.verbindung);
		factory.registerStandardFactories();
	}

	/**
	 * Prefix der PID der zu erzeugenden Meldung.
	 */
	private final String meldungsPidPrefix = "rdsmeldungtest.";

	/**
	 * Landesmeldestelle.
	 */
	private final String landesmeldeStelle = "lmst.bw";

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		// TODO Auto-generated method stub

	}

	/**
	 * Erzeugt eine RDS-Meldung und versendet diese an den DaV.
	 */
	@Test
	public void erzeugeMeldung() {

		final SystemObject lms = ErzeugeMeldung.verbindung.getDataModel().getObject(landesmeldeStelle);
		SystemObject rdsmeldung = null;

		try {
			final MutableSet meldungsmenge = ((ConfigurationObject) lms).getMutableSet("RDSMeldungen");

			final DataModel model = ErzeugeMeldung.verbindung.getDataModel();

			final String meldungsPid = meldungsPidPrefix + String.valueOf(System.currentTimeMillis());

			rdsmeldung = model.getObject(meldungsPid);
			if (rdsmeldung == null) {
				final ConfigurationArea area = ErzeugeMeldung.verbindung.getLocalConfigurationAuthority()
						.getConfigurationArea();
				System.out.println("Die neue Meldung '" + meldungsPid + "' wird angelegt");
				rdsmeldung = area.createDynamicObject((DynamicObjectType) model.getType("typ.rdsMeldung"), meldungsPid,
						"Testmeldung " + meldungsPid);
				meldungsmenge.add(rdsmeldung);
			}

			System.out.println("Sende Daten für die Meldung '" + rdsmeldung + "', PID: " + rdsmeldung.getPid());

			DavDatenModell.getInstanz().setVerbindung(ErzeugeMeldung.verbindung);
			final RDSMeldung m = new RDSMeldung(rdsmeldung);

			final DataDescription dd = new DataDescription(model.getAttributeGroup("atg.rdsMeldung"),
					model.getAspect("asp.rdsSenden"));

			ErzeugeMeldung.verbindung.subscribeSender(this, m.getSystemObject(), dd, SenderRole.source());

			final Data data = ErzeugeMeldung.verbindung.createData(dd.getAttributeGroup());
			m.getMeldungsDaten().publizieren(data);
			final ResultData meldungsdaten = new ResultData(m.getSystemObject(), dd, System.currentTimeMillis(), data);
			ErzeugeMeldung.verbindung.sendData(meldungsdaten);

			// etwas warten, damit Quelle nicht verschwindet
			Thread.sleep(200000);
		} catch (final Exception e) {
			e.printStackTrace();
			Assert.fail(e.getMessage());
		}
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		// TODO Auto-generated method stub
		return false;
	}
}
