/*
 * Segment 2 (KEx),  SWE 2.1 LMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lms.vew;

import java.util.logging.Level;
import java.util.logging.Logger;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.lms.empfang.dekoder.Meldungsdekodierer;
import de.bsvrz.kex.lms.versand.kodierer.Meldungskodierer;
import de.bsvrz.kex.lms.vew.daten.DavDatenModell;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Hauptklasse f&uuml;r die SWE 2.1 LMS.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class KexLMS implements StandardApplication {

	/**
	 * Landesmeldestellen-Objekt.
	 */
	private SystemObject lmsobjekt;

	/**
	 * Worker-Thread.
	 */
	private class KexLMSWorker extends Thread {

		@Override
		public void run() {
			Meldungskodierer mk = null;
			Meldungsdekodierer md = null;
			String fehler = null;

			LogTools.log(debug, LmsMsg.InfoApplikationStartet);

			DavDatenModell.getInstanz().setVerbindung(dav);

			try {
				// Komponente Meldungsversand starten
				mk = new Meldungskodierer(dav, (ConfigurationObject) lmsobjekt);

				// ohne diese Pause geht der Meldungsdekodierer oft in Fehler:
				// "Die
				// Menge 'RDSMeldungen' existiert nicht am Objekt..."
				Thread.sleep(5000);

				// Komponente Meldungsempfang starten
				md = new Meldungsdekodierer(dav, (ConfigurationObject) lmsobjekt);

				while (true) {
					Thread.sleep(10000);
					if (!mk.isAlive()) {
						// LogTools.log(debug, LmsMsg.ErrorThreadBeendet,
						// "Meldungskodierer");
						md.stopp();
						fehler = "Meldungskodierer-Thread beendet";
						break;
					}

					if (!md.isAlive()) {
						// LogTools.log(debug, LmsMsg.ErrorThreadBeendet,
						// "Meldungsdekodierer");
						fehler = "Meldungsdekodierer-Thread beendet";
						mk.stopp();
						break;
					}
				}
			} catch (final Exception e) {
				debug.error(e.getMessage());
			}

			if ((mk != null) && mk.isAlive()) {
				mk.stopp();
				try {
					mk.join();
				} catch (final InterruptedException e) {
				}
			}

			if ((md != null) && md.isAlive()) {
				md.stopp();
				try {
					mk.join();
				} catch (final InterruptedException e) {
				}
			}

			if (fehler != null) {
				LogTools.log(debug, LmsMsg.ErrorAppBeendet, fehler);
			} else {
				LogTools.log(debug, LmsMsg.InfoApplikationBeendet);
			}
		}
	}

	/**
	 * MAIN-Methode.
	 *
	 * @param args
	 *            Argumente
	 */
	public static void main(final String[] args) {
		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {

			@Override
			public void uncaughtException(final Thread t, final Throwable e) {
				Logger.getLogger(getClass().getName()).log(Level.INFO, "*** Application wird beendet ***", e); //$NON-NLS-1$
			}

		});

		StandardApplicationRunner.run(new KexLMS(), args);
	}

	/**
	 * DebugLogger für Debug-Ausgaben.
	 */
	Debug debug;

	/**
	 * Die Verbindung zum Datenverteiler.
	 */
	ClientDavConnection dav;

	/**
	 * Standard-Konstruktor.
	 */
	protected KexLMS() {
		super();
	}

	/**
	 * &Uuml;berpr&uuml;ft, ob alle notwendigen Parameter f&uuml;r die
	 * Applikation gesetzt sind.
	 *
	 * @throws RDSException
	 *             bei fehlenden notwendigen Paramertern
	 */
	public void checkParameter() throws RDSException {
		// teste notwendige Parameter
		final String[] test = { LMSProperties.PROP_KEY_LMS };

		for (final String element : test) {
			final String testprop = LMSProperties.getInstance().getProperty(element);
			if ((testprop == null) || (testprop.length() == 0)) {
				throw new RDSException("Der Parameter '" + element + "' fehlt!"); //$NON-NLS-1$//$NON-NLS-2$
			}
		}
	}

	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {
		debug = Debug.getLogger();
		dav = (ClientDavConnection) connection;

		try {
			checkParameter();

			// teste, ob das Landesmeldestellen-Objekt existiert
			final String lms = LMSProperties.getInstance().getProperty(LMSProperties.PROP_KEY_LMS);
			if ((lms == null) || (lms.length() == 0)) {
				throw new RDSException("Der Parameter 'Landesmeldestelle' ist nicht gesetzt!");
			}

			lmsobjekt = dav.getDataModel().getObject(lms);
			if (lmsobjekt == null) {
				throw new RDSException("Das Objekt '" + lms + "' kann nicht vom Datenverteiler gelesen werden!");
			}

			new KexLMSWorker().start();
		} catch (final Exception e) {
			LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
			throw (e);
		}
	}

	/**
	 * Wertet die an die Applikation &uuml;bergebenen Kommandozeilenparameter
	 * aus.<br>
	 * Zus&auml;tzlich zu den Standard-Datenverteiler-Argumenten sind folgende
	 * Argumente m&ouml;glich:
	 * <ul>
	 * <li>-LMS=&lt;Konfigurationsobjekt&gt;, PID oder ID der Landesmeldestelle,
	 * mit der Meldungen ausgetauscht werden sollen</li>
	 * <li>-versandDateiPrefix=&lt;Text&gt;, Legt den Pr&auml;fix der Dateinamen
	 * der an die LMS zu versendenden Dateien fest Standard: vrzrds</li>
	 * <li>-sicherungsVerzeichnis=&lt;Pfad&gt;, Legt das Dateiverzeichnis fest,
	 * in das Meldungsdateien kopiert werden, bei deren Auswertung Fehler
	 * aufgetreten sind. Standard: keine Sicherung</li>
	 * <li>-sicherungsZeit=&lt;Tage&gt;, Anzahl Tage, die Dateien im
	 * Sicherungsverzeichnis gespeichert werden Standard: 5 Tage</li>
	 * </ul>
	 *
	 * @param argumentList
	 *            die Liste der &uuml;bergebenen Argumente
	 * @throws Exception
	 *             bei Ausnahmen
	 * @see de.bsvrz.sys.funclib.application.StandardApplication#parseArguments(de.bsvrz.sys.funclib.commandLineArgs.ArgumentList)
	 */
	@Override
	public void parseArguments(final ArgumentList argumentList) throws Exception {
		final String[] myargs = { LMSProperties.PROP_KEY_LMS, LMSProperties.PROP_KEY_SICHERUNGSVERZEICHNIS,
				LMSProperties.PROP_KEY_SICHERUNGSZEIT, LMSProperties.PROP_KEY_VERSANDDATEIPREFIX };

		for (final String myarg : myargs) {
			if (argumentList.hasArgument(myarg)) {
				LMSProperties.getInstance().setProperty(myarg, argumentList.fetchArgument(myarg).asString());
			}
		}
	}
}
