/*
 * Segment 2 (KEx),  SWE 2.1 LMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lms.vew;

import java.io.File;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.resource.Configuration;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.lms.MeldungenLoeschen;
import de.bsvrz.kex.lms.empfang.dekoder.TICInfoDatei;
import de.bsvrz.kex.lms.empfang.dekoder.TICInfoDateiException;
import de.bsvrz.kex.lms.vew.daten.DavDatenModell;
import de.bsvrz.kex.lms.vew.daten.RDSMeldungsDaten;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;

/**
 * Erzeugt eine RDS-Meldung im Datenverteiler.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class VersendeMeldung implements ClientSenderInterface {
	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(MeldungenLoeschen.class.getName());

	private static ClientDavConnection verbindung;

	private static Configuration konfig;

	/**
	 * @throws Exception
	 */
	@BeforeClass
	public static void beforeClass() throws Exception {
		final ObjektFactory factory;

		final Level level;

		VersendeMeldung.konfig = Configuration.getConfiguration();
		level = Level.parse(VersendeMeldung.konfig.getString("debug.level"));
		LoggerTools.setLoggerLevel(level);
		VersendeMeldung.log.setLevel(level);

		VersendeMeldung.log.info("Stelle Verbindung zum Datenverteiler her ...");
		VersendeMeldung.verbindung = new ClientDavConnection();
		VersendeMeldung.verbindung.getClientDavParameters()
		.setDavCommunicationAddress(VersendeMeldung.konfig.getString("dav.host"));
		VersendeMeldung.verbindung.getClientDavParameters()
		.setDavCommunicationSubAddress(VersendeMeldung.konfig.getInt("dav.port"));
		VersendeMeldung.verbindung.connect();
		VersendeMeldung.verbindung.login(VersendeMeldung.konfig.getString("dav.benutzer"),
				VersendeMeldung.konfig.getString("dav.kennwort"));

		factory = ObjektFactory.getInstanz();
		factory.setVerbindung(VersendeMeldung.verbindung);
		factory.registerStandardFactories();
	}

	/**
	 * Prefix der PID der zu erzeugenden Meldung.
	 */
	private final String meldungsPidPrefix = "rdsmeldungtest.";

	/**
	 * Landesmeldestelle.
	 */
	private final String landesmeldeStelle = "lmst.bw";

	private List<RDSMeldung> bildeMeldungenAusDatei(final String dateiname) throws TICInfoDateiException, RDSException {
		final TICInfoDatei tid = new TICInfoDatei(new File(dateiname));
		final List<RDSMeldung> meldungen = new ArrayList<RDSMeldung>();

		for (final RDSMeldungsDaten meldungsdaten : tid.bildeMeldungen()) {
			meldungen.add(neueMeldung(meldungsdaten));
		}
		return meldungen;
	}

	private List<RDSMeldung> bildeMeldungenAusDateien() throws TICInfoDateiException, RDSException {
		final List<RDSMeldung> meldungen = new ArrayList<RDSMeldung>();
		final String dateiliste = VersendeMeldung.konfig.getString("VersandMeldung");

		for (final String datei : dateiliste.split(",")) {
			meldungen.addAll(bildeMeldungenAusDatei(datei));
		}

		return meldungen;
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		// TODO Auto-generated method stub

	}

	/**
	 * Erzeugt eine RDS-Meldung und versendet diese an den DaV.
	 */
	@Test
	public void erzeugeMeldungen() {

		final SystemObject lms = VersendeMeldung.verbindung.getDataModel().getObject(landesmeldeStelle);

		try {
			final DataDescription dd = new DataDescription(
					VersendeMeldung.verbindung.getDataModel().getAttributeGroup("atg.rdsMeldung"),
					VersendeMeldung.verbindung.getDataModel().getAspect("asp.rdsSenden"));

			DavDatenModell.getInstanz().setVerbindung(VersendeMeldung.verbindung);

			final MutableSet meldungsmenge = ((ConfigurationObject) lms).getMutableSet("RDSMeldungen");

			final DateFormat formatter = new SimpleDateFormat();
			final GregorianCalendar cal = new GregorianCalendar();

			for (final RDSMeldung meldung : bildeMeldungenAusDateien()) {
				meldungsmenge.add(meldung.getSystemObject());

				System.out.println(
						"Sende Daten für die Meldung '" + meldung + "', PID: " + meldung.getSystemObject().getPid());

				VersendeMeldung.verbindung.subscribeSender(this, meldung.getSystemObject(), dd, SenderRole.source());

				final Data data = VersendeMeldung.verbindung.createData(dd.getAttributeGroup());

				final long dauer = Long.parseLong(VersendeMeldung.konfig.getString("Gueltigkeitsdauer"));

				cal.setTimeInMillis(System.currentTimeMillis() + dauer);

				System.out.println("Ablaufzeit:       " + formatter.format(cal.getTime()));
				meldung.getMeldungsDaten().setAblaufzeit(cal.getTimeInMillis());

				meldung.getMeldungsDaten().publizieren(data);
				final ResultData meldungsdaten = new ResultData(meldung.getSystemObject(), dd,
						System.currentTimeMillis(), data);
				VersendeMeldung.verbindung.sendData(meldungsdaten);
			}

			// bis zum Ende der Gültigkeit warten, damit Quelle nicht
			// verschwindet
			Thread.sleep(cal.getTimeInMillis() - System.currentTimeMillis());
		} catch (final Exception e) {
			e.printStackTrace();
			Assert.fail(e.getMessage());
		}
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		// TODO Auto-generated method stub
		return false;
	}

	private RDSMeldung neueMeldung(final RDSMeldungsDaten meldungsdaten) throws RDSException {
		final ConfigurationArea area = VersendeMeldung.verbindung.getLocalConfigurationAuthority()
				.getConfigurationArea();

		final String pid = meldungsdaten.bildePid();
		SystemObject so = VersendeMeldung.verbindung.getDataModel().getObject(pid);
		if (so == null) {
			try {
				so = area.createDynamicObject(
						(DynamicObjectType) VersendeMeldung.verbindung.getDataModel().getType("typ.rdsMeldung"), pid,
						pid);
				System.out.println("Die Meldung " + pid + " wurde angelegt");
			} catch (final ConfigurationChangeException e) {
				throw new RDSException(e.getMessage());
			}
		} else {
			System.out.println("Die Meldung " + pid + " existiert bereits");
		}

		final RDSMeldung rdsmeldung = new RDSMeldung(so, meldungsdaten);

		return rdsmeldung;
	}
}
