/*
 * Segment 2 (KEx),  SWE 2.1 LMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lms.vew.daten;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.kex.lms.vew.RDSException;

/**
 * Nachrichtentexte einer RDS-Meldung.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class RDSNachrichten extends AbstractRDSInfoElement {
	/**
	 * Nachrichtendaten.
	 */
	class RDSNachrichtenDaten extends AbstractRDSInfoElement {

		/**
		 * TIC Info XML Element: Message, enthält alle textuellen Nachrichten
		 * zur Meldung.
		 */
		protected static final String XML_ELEMENT_MESSAGEDATA = "MDA";

		/**
		 * Klassifikation des Textes.
		 */
		// RDS_Nachrichtenklasse _nachrichtenklasse =
		// RDS_Nachrichtenklasse.freier_Text;
		/**
		 * TIC Info XML Element: MessageDataClass, klassifiziert den Text.
		 */
		private static final String XML_ELEMENT_MESSAGEDATACLASS = "MDC";

		/**
		 * Text der Nachricht
		 */
		// String _nachrichtenText = "nicht gesetzt";
		/**
		 * TIC Info XML Element: MessageDataText, Text.
		 */
		private static final String XML_ELEMENT_MESSAGEDATATEXT = "MDT";

		/**
		 * Erzeugt ein neues Element.
		 */
		public RDSNachrichtenDaten() {
			super(RDSNachrichtenDaten.XML_ELEMENT_MESSAGEDATA, "");
			appendChild(new RDSInteger(RDSNachrichtenDaten.XML_ELEMENT_MESSAGEDATACLASS, "NachrichtenKlasse"));
			appendChild(new RDSText(RDSNachrichtenDaten.XML_ELEMENT_MESSAGEDATATEXT, "NachrichtenText"));
		}
	}

	/**
	 * TIC Info XML Element: Message, enthält alle textuellen Nachrichten zur
	 * Meldung.
	 */
	private static final String XML_ELEMENT_MESSAGE = "MES";

	/**
	 * TIC Info XML Element: MessageLangID, spezifiziert die Sprache der
	 * Nachrichten entsprechend TIC Info Specification: Anhang 'Language
	 * Identifier'.
	 */
	private static final String XML_ELEMENT_MESSAGELANGID = "MDL";
	/**
	 * Liste der Nachrichtendaten.
	 */
	List<RDSNachrichtenDaten> nachrichtenDaten = new ArrayList<RDSNachrichtenDaten>();

	/**
	 * Sprachkennung der Nachrichtentexte.
	 */
	int nachrichtenSprache = 1031; // deutsch (Standard)" wert="1031

	/**
	 * Erzeugt ein neues Objekt.
	 *
	 * @param davattribut
	 *            Name des zugeh&ouml;rigen DaV-Attributes
	 */
	public RDSNachrichten(final String davattribut) {
		super(RDSNachrichten.XML_ELEMENT_MESSAGE, davattribut);
		appendChild(new RDSZustand(RDSNachrichten.XML_ELEMENT_MESSAGELANGID, "NachrichtenSprache"));
	}

	@Override
	public void aktualisieren(final Data davdata) throws RDSException {
		super.aktualisieren(davdata);
		final Data data = davdata.getItem(getDavAttributName());
		final Data.Array array = data.getArray("NachrichtenDaten");
		final int anzahl = array.getLength();
		nachrichtenDaten.clear();
		for (int i = 0; i < anzahl; i++) {
			final RDSNachrichtenDaten nd = new RDSNachrichtenDaten();
			nd.aktualisieren(array.getItem(i));
			nachrichtenDaten.add(nd);
		}
	}

	@Override
	public void einlesenXML(final Node node) throws RDSException {
		super.einlesenXML(node);
		final NodeList nl = node.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			final Node n = nl.item(i);
			if (n.getNodeName().equals(RDSNachrichtenDaten.XML_ELEMENT_MESSAGEDATA)) {
				final RDSNachrichtenDaten nd = new RDSNachrichtenDaten();
				nd.einlesenXML(n);
				nachrichtenDaten.add(nd);
				// appendChild(nd);
			}
		}
	}

	@Override
	public Element erzeugeDocumentElement(final Document doc, final Element parent) throws RDSException {
		final Element oelement = super.erzeugeDocumentElement(doc, parent);

		for (final RDSNachrichtenDaten d : nachrichtenDaten) {
			d.erzeugeDocumentElement(doc, oelement);
		}

		return oelement;
	}

	@Override
	public void publizieren(final Data davdata) throws RDSException {
		super.publizieren(davdata);
		final Data data = davdata.getItem(getDavAttributName());

		final Data.Array array = data.getArray("NachrichtenDaten");
		final int anzahl = nachrichtenDaten.size();
		array.setLength(anzahl);
		for (int i = 0; i < anzahl; i++) {
			nachrichtenDaten.get(i).publizieren(array.getItem(i));
		}
	}
}
